"""
-----------------------------------------------------------------------
                       Poisson in a 3D QPC device
-----------------------------------------------------------------------

(c) 2017 Adriaan Vuik (TU Delft).

This module contains functions to build a FEM mesh of a 3D QPC nanowire 
device and to solve Poisson's equation on this mesh, give a charge density.
"""


import numpy as np
import logging
logging.getLogger('FFC').setLevel(logging.WARNING)
import dolfin
import mshr
dolfin.parameters['allow_extrapolation'] = True

    

def refine_mesh(mesh, subdomain):
    cc = dolfin.CellFunction("bool", mesh, False)
    subdomain.mark(cc, True)
    mesh = dolfin.refine(mesh, cc)
    return mesh


def wire_coords(mesh, dim_par):
    a0 = 0.5*dim_par.radius
    b0 = np.sin(np.pi/3.0)*dim_par.radius
    mesh_points_all = mesh.coordinates()
    num_nodes = len(mesh_points_all)
    
    x = mesh_points_all[:, 0]
    y = mesh_points_all[:, 1]
    z = mesh_points_all[:, 2]
    mesh_points_filter_x = []
    mesh_points_filter_y = []
    mesh_points_filter_z = []
    
    for i in range(num_nodes):
        if (y[i] >= -b0 and y[i] <= b0 and 
            y[i] >= - (b0/a0)*(2*a0 - x[i]) and 
            y[i] <= - (b0/a0)*(x[i] - 2*a0) and 
            y[i] <= (b0/a0)*(x[i] + 2*a0) and 
            y[i] >= - (b0/a0)*(x[i] + 2*a0)):
            mesh_points_filter_x.append(x[i])
            mesh_points_filter_y.append(y[i])
            mesh_points_filter_z.append(z[i])

    mesh_points = list(zip(mesh_points_filter_x, mesh_points_filter_y, mesh_points_filter_z))
    return mesh_points, num_nodes


def build_mesh(mesh_refinement, dim_par):
    """Make a 3d mesh of the nanowire and surrounding electrostatic components"""
    # Unpack all dimensions
    radius = dim_par.radius
    d_dielec = dim_par.d_dielec
    length = dim_par.L
    sc_length = dim_par.sc_length
    normal_length = dim_par.normal_length
    a0 = 0.5*radius
    b0 = np.sin(np.pi/3.0)*radius
    bottom = d_dielec + b0
    top = dim_par.top
    side = 0.5*dim_par.width
    
    # Combine the dimensions into closed shapes
    outer_boundary = mshr.Box(dolfin.Point(-side, -b0, 0),
                              dolfin.Point(side, top, length))
    hexa_wire =  mshr.Cylinder(dolfin.Point(0, 0, 0), dolfin.Point(0, 0, length), 
                               radius, radius, 6)
    dielec_layer = mshr.Box(dolfin.Point(-side, -bottom, 0),
                            dolfin.Point(side, -b0 + 1e-3, length))
    normal_lead1 = mshr.Cylinder(dolfin.Point(-3*a0, b0, 0), dolfin.Point(-3*a0, b0, normal_length),
                               radius + 1e-3, radius + 1e-3, 6)
    normal_lead2 = mshr.Cylinder(dolfin.Point(3*a0, b0, 0), dolfin.Point(3*a0, b0, normal_length),
                               radius + 1e-3, radius + 1e-3, 6)
    normal_lead3 = mshr.Cylinder(dolfin.Point(0, 2*b0, 0), dolfin.Point(0, 2*b0, normal_length),
                               radius + 1e-3, radius + 1e-3, 6)
    sc_lead1 = mshr.Cylinder(dolfin.Point(-3*a0, b0, length-sc_length), dolfin.Point(-3*a0, b0, length),
                               radius + 1e-3, radius + 1e-3, 6)
    sc_lead2 = mshr.Cylinder(dolfin.Point(3*a0, b0, length-sc_length), dolfin.Point(3*a0, b0, length),
                               radius + 1e-3, radius + 1e-3, 6)
    sc_lead3 = mshr.Cylinder(dolfin.Point(0, 2*b0, length-sc_length), dolfin.Point(0, 2*b0, length),
                               radius + 1e-3, radius + 1e-3, 6)
    gate = mshr.Box(dolfin.Point(-side, -bottom, 0), dolfin.Point(side, -bottom - 0.1*b0, length))
        
    geometry = (dielec_layer + hexa_wire + normal_lead1 + normal_lead2 + normal_lead3 +
                sc_lead1 + sc_lead2 + sc_lead3 + outer_boundary + gate)
    
               
    mesh = mshr.generate_mesh(geometry, mesh_refinement)
    
    buf = 0.1*radius
    hexagon = dolfin.AutoSubDomain(lambda x:
                    -b0 - buf <= x[1] <= b0 + buf  and
                    x[1] >= -(b0/a0)*(2*a0 - x[0]) - buf  and
                    x[1] <= -(b0/a0)*(x[0] - 2*a0) + buf and
                    x[1] <= (b0/a0)*(2*a0 + x[0]) + buf  and
                    x[1] >= -(b0/a0)*(x[0] + 2*a0) - buf )
    dielec = dolfin.AutoSubDomain(lambda x: -bottom <= x[1] < -b0 + 0.1*d_dielec)
    super_lead = dolfin.AutoSubDomain(lambda x:
                    (-a0*x[1]/b0 - 2.5*a0 < x[0] <= -a0*x[1]/b0 - 2*a0 and -b0 < x[1] <= 0) or 
                    (a0*x[1]/b0 - 2.5*a0 < x[0] <= a0*x[1]/b0 - 2*a0 and 0 < x[1] <= b0) or
                    (-2*a0 < x[0] < 2*a0 and b0 < x[1] <= top) or
                    (-a0*x[1]/b0 + 2.5*a0 > x[0] >= -a0*x[1]/b0 + 2*a0 and 0 <= x[1] < b0) or
                    (a0*x[1]/b0 + 2.5*a0 > x[0] >= a0*x[1]/b0 + 2*a0 and -b0 < x[1] <= 0) and
                     length - sc_length <= x[2])
    normal_lead = dolfin.AutoSubDomain(lambda x:
                    (-a0*x[1]/b0 - 2.5*a0 < x[0] <= -a0*x[1]/b0 - 2*a0 and -b0 < x[1] <= 0) or 
                    (a0*x[1]/b0 - 2.5*a0 < x[0] <= a0*x[1]/b0 - 2*a0 and 0 < x[1] <= b0) or
                    (-2*a0 < x[0] < 2*a0 and b0 < x[1] <= top) or
                    (-a0*x[1]/b0 + 2.5*a0 > x[0] >= -a0*x[1]/b0 + 2*a0 and 0 <= x[1] < b0) or
                    (a0*x[1]/b0 + 2.5*a0 > x[0] >= a0*x[1]/b0 + 2*a0 and -b0 < x[1] <= 0) and
                     x[2] <= normal_length)
    
    gate = dolfin.AutoSubDomain(lambda x: -bottom <= x[1] < -0.7*bottom )
    dielec_interface = dolfin.AutoSubDomain(lambda x: -b0 - 0.3*bottom < x[1] < -b0 + 0.3*radius)
    contact_interface = dolfin.AutoSubDomain(lambda x: 0.7*radius <= x[1] <= 1.3*radius)
    
    subdom_list = [hexagon, dielec]
    subdom_list2 = [normal_lead, super_lead]
    for subdom in subdom_list:
        for i in range(2):
            mesh = refine_mesh(mesh, subdom) 
    for subdom in subdom_list2:
        for i in range(1):
            mesh = refine_mesh(mesh, subdom)
                
    return mesh


def boundary_regions(dim_par, discr_par):
    """Makes the boundary regions of the mesh in order to specify 
       boundary conditions."""
    radius = dim_par.radius
    d_dielec = dim_par.d_dielec
    sc_length = dim_par.sc_length
    normal_length = dim_par.normal_length
    L = dim_par.L
    a0 = 0.5*radius
    b0 = np.sin(np.pi/3.0)*radius
    bottom = d_dielec + b0
    top = dim_par.top
    side = 0.5*dim_par.width
    
    V = discr_par.V
    mesh = discr_par.mesh

    class Plate(dolfin.SubDomain):
        def inside(self, x, on_boundary):
            return (x[1] <= -bottom)

    class Wire(dolfin.SubDomain):
        def inside(self, x, on_boundary):
            return ((-b0 <= x[1] <= b0) and
                    (-(b0/a0)*(2*a0 - x[0]) <= x[1] <= (b0/a0)*(2*a0 - x[0])) and
                    (-(b0/a0)*(2*a0 + x[0]) <= x[1] <= (b0/a0)*(x[0] + 2*a0)))

    class Dielec(dolfin.SubDomain):
        def inside(self, x, on_boundary):
            return (-bottom < x[1] <= -b0)
        
    class Superconducting_Lead(dolfin.SubDomain):
        def inside(self, x, on_boundary):
            return (((x[0] <= -a0*x[1]/b0 - 2*a0 and -b0 < x[1] <= 0) or 
                    (x[0] <= a0*x[1]/b0 - 2*a0 and 0 < x[1] <= b0) or
                    (b0 < x[1] <= top) or
                    (x[0] >= -a0*x[1]/b0 + 2*a0 and 0 <= x[1] < b0) or
                    (x[0] >= a0*x[1]/b0 + 2*a0 and -b0 < x[1] <= 0)) and
                     L - sc_length <= x[2])
        
        
    class Normal_Lead(dolfin.SubDomain):
        def inside(self, x, on_boundary):
            return (((x[0] <= -a0*x[1]/b0 - 2*a0 and -b0 < x[1] <= 0) or 
                    (x[0] <= a0*x[1]/b0 - 2*a0 and 0 < x[1] <= b0) or
                    (b0 <= x[1] <= top) or
                    (x[0] >= -a0*x[1]/b0 + 2*a0 and 0 < x[1] <= b0) or
                    (x[0] >= a0*x[1]/b0 + 2*a0 and -b0 < x[1] <= 0)) and
                     x[2] <= normal_length)
                              

    plate = Plate()
    wire = Wire()
    dielec = Dielec()
    sc_lead = Superconducting_Lead()
    normal_lead = Normal_Lead()

    domains = dolfin.CellFunction("size_t", mesh)
    domains.set_all(0)
    dielec.mark(domains, 1)
    wire.mark(domains, 2)

    boundaries = dolfin.FacetFunction("size_t", mesh)
    boundaries.set_all(0)
    plate.mark(boundaries, 1)
    sc_lead.mark(boundaries, 2)
    normal_lead.mark(boundaries, 3)

    dx = dolfin.Measure("dx")(subdomain_data=domains)
    ds = dolfin.Measure("ds")(subdomain_data=boundaries)
    
    return boundaries, dx, ds


def poisson_potential(density, bound_conds, c, dim_par, discr_par):
    V_G = bound_conds.V_G
    V_NL = bound_conds.V_NL
    V_SC = bound_conds.V_SC
    V = discr_par.V
    
    boundaries, dx, ds = boundary_regions(dim_par, discr_par)
    
    avac = dolfin.Constant(1.0)
    a1 = c.eps_dielec
    a2 = c.eps_wire

    u = dolfin.TrialFunction(V)
    v = dolfin.TestFunction(V)
    a = (dolfin.inner(avac*dolfin.grad(u), dolfin.grad(v))*dx(0)
        +dolfin.inner(a1*dolfin.grad(u), dolfin.grad(v))*dx(1)
        +dolfin.inner(a2*dolfin.grad(u), dolfin.grad(v))*dx(2))
    L = density*v*dx(2)
    
    bcs = [dolfin.DirichletBC(V, V_G, boundaries, 1),
            dolfin.DirichletBC(V, V_NL, boundaries, 2),
            dolfin.DirichletBC(V, V_SC, boundaries, 3)]

    u = dolfin.Function(V)
    dolfin.solve(a == L, u, bcs)
    poisson_pot = u.vector()[:]

    return poisson_pot