"""
-----------------------------------------------------------------------
                         Thomas-Fermi in 3D
-----------------------------------------------------------------------

(c) 2017 Adriaan Vuik (TU Delft).

This module allows to compute the Poisson equation in three-dimensional systems
using the Thomas-Fermi method as an approximation for the charge density.
"""

import numpy as np
from scipy import optimize, interpolate
import logging
logging.getLogger('FFC').setLevel(logging.WARNING)
import dolfin

import QPC_device as qpcd


### Constants ###
hbar = 1.055e-34 
m = 9.109e-31 
eV = 1.6023e-19
e = 1.6023e-19
eps0 = 8.854e-12 
k_B = 1.380648e-23

# Indium Antimonide (InSb)
eps_InSb = 17.7 
m_eff_InSb = 0.014*m 
bandgap_InSb = 0.1 #half the bandgap of InSb (in eV)

# Silicon Nitride (Si3N4)
eps_SiN = 8.0


def reorder_source_coordinates(poisson_density, discr_par):
    V = discr_par.V
    mesh = discr_par.mesh
    mesh_points = discr_par.mesh_points
    if (mesh.geometry().dim() == 2):
        to_1d = lambda x: x.view(dtype=np.dtype([('f1', 'float'),
                                                 ('f2', 'float')]))[:, 0]
    if (mesh.geometry().dim() == 3):
        to_1d = lambda x: x.view(dtype=np.dtype([('f1', 'float'),
                                                 ('f2', 'float'),
                                                 ('f3', 'float')]))[:, 0]
    
    n,d = V.dim(), mesh.geometry().dim()
    F = dolfin.FunctionSpace(mesh, "Lagrange", 1)
    V_dof_coordinates = np.array(F.tabulate_dof_coordinates())
    V_dof_coordinates.resize((n,d))
    ordering = np.argsort(to_1d(V_dof_coordinates))
    wire_indices = np.searchsorted(to_1d(V_dof_coordinates),
                                   to_1d(np.array(mesh_points)), sorter=ordering)
    poisson_density_reordered = dolfin.Function(V)
    poisson_density_reordered.vector()[ordering[wire_indices]] = poisson_density
    return poisson_density_reordered


def refine_gridpoint(x, x_range, n):
    """Refine a gridpoint x in a grid range x_range to n substeps
    between x and x + delta_x"""
    delta_x = abs(x_range[1] - x_range[0])
    delta_x_range = np.linspace(x, x+delta_x, n + 1)[:n]
    return delta_x_range


def refine_grid(x_range, n):
    """Refine all gridpoints in x_range to n substeps, returning a refined range
    of n times len(x_range) gridpoints"""
    refined_range = [refine_gridpoint(x, x_range, n) for x in x_range]
    refined_range = np.reshape(np.array(refined_range), (n*len(x_range), 1))
    return refined_range


def thomas_fermi_density(u, E_bandgap, discr_par):
    """Computes the charge density according to the Thomas-Fermi approximation.
    Cut off whenever phi(x, y) < 0."""
    mesh_points = discr_par.mesh_points
    tf_scaling = (-e*(discr_par.a*1e-9)**2/eps0)*\
                 (1/(3*np.pi**2))*(2*e*m_eff_InSb/hbar**2)**(3/2.)
   
    pot = []
    for coord in mesh_points:
        pot.append(u(dolfin.Point(*coord)))
    # Subtract the bandgap in units of eV
    pot_bandgap = np.subtract(pot, 1e-3*E_bandgap) 
    pot_bandgap[pot_bandgap < 0] = 0
    density = tf_scaling * pot_bandgap**(3/2.)
    density_function = reorder_source_coordinates(np.array(density), discr_par)
    return density_function


def thomas_fermi_approximation(bound_conds, c, dimension_par,  discr_par):
    """Solves the Poisson equation using the Thomas-Fermi density on the 
    right-hand side. The method uses a simple Picard iteration to solve 
    self-consistently."""
    eps_vac = dolfin.Constant(1.0)
    eps_wire = c.eps_wire
    eps_dielec = c.eps_dielec
    E_bandgap = c.E_bandgap
    Ez = c.Ez
    Eso = c.Eso
    
    V = discr_par.V
    
    boundaries, dx, ds = nsj.boundary_regions(dimension_par, discr_par)
    bcs = [dolfin.DirichletBC(V, bound_conds.V_G, boundaries, 1),
               dolfin.DirichletBC(V, bound_conds.V_NL, boundaries, 2),
               dolfin.DirichletBC(V, bound_conds.V_SC, boundaries, 3)]
    
    # Define variational problem for Picard iteration
    u = dolfin.TrialFunction(V)
    v = dolfin.TestFunction(V)

    init_guess = 0.1*(bound_conds.V_G + 3*bound_conds.V_SC + 3*bound_conds.V_NL)
    u_k = dolfin.interpolate(dolfin.Constant(init_guess), V)  # previous (known) u
    a = (dolfin.inner(eps_vac*dolfin.grad(u), dolfin.grad(v))*dx(0)
         +dolfin.inner(eps_dielec*dolfin.grad(u), dolfin.grad(v))*dx(1)
         +dolfin.inner(eps_wire*dolfin.grad(u), dolfin.grad(v))*dx(2))

    # Picard iterations
    u = dolfin.Function(V) 
    eps = 1.0           # error measure ||u-u_k||
    tol = 1.0E-4       # tolerance
    iter = 0            # iteration counter
    maxiter = 4000        # max no of iterations allowed
    while eps > tol and iter < maxiter:
        L = thomas_fermi_density(u_k, E_bandgap, discr_par)*v*dx(2)
        iter += 1
        dolfin.solve(a == L, u, bcs)
        diff = u.vector().array() - u_k.vector().array()
        eps = np.linalg.norm(diff, ord=np.Inf)
        #print eps
        # damp the new solution by mixing with the previous one
        u_mix = dolfin.Function(V)
        u_mix.vector()[:] = (0.05*u.vector()[:] + 0.95*u_k.vector()[:])
        u_k.assign(u_mix)   

    return u


def thomas_fermi_anderson(bound_conds, c, dim_par,  discr_par):
    """Solves the Poisson equation using the Thomas-Fermi density on the 
    right-hand side. The method uses the Anderson mixing scheme to solve 
    self-consistently."""
    eps_vac = dolfin.Constant(1.0)
    eps_wire = c.eps_wire
    eps_dielec = c.eps_dielec
    E_bandgap = c.E_bandgap
    Ez = c.Ez
    Eso = c.Eso
    
    V = discr_par.V

    
    def f_minimize_pot(potential):
        pot_func = dolfin.Function(V)
        pot_func.vector()[:] = potential
        density_func = thomas_fermi_density(pot_func, E_bandgap, discr_par)
        potential_new = qpcd.poisson_potential(density_func, bound_conds, c, dim_par, discr_par)
        F = potential_new - potential
        print('error is ', np.linalg.norm(F))
        return F  

    return f_minimize_pot


def tf_potential_anderson_iterative(V_G, init_guess, bound_conds, c, dim_par, discr_par, solver_kwargs):
    """Computes the TF potential using Anderson optimization. Initial guess must be provided"""
    bound_conds.V_G = V_G
    optimize_func = thomas_fermi_anderson(bound_conds, c, dim_par, discr_par)
    tf_pot_arr = optimize.anderson(optimize_func, init_guess, **solver_kwargs)
    return tf_pot_arr


def tf_pot_gate_refining(V_G, V_G_range, n_refine, bound_conds, c, dim_par, discr_par, solver_kwargs):
    """Function to be used when a limited number of engines is available for parallel computation.
    It takes a range of potential values, refines around each value, and computes the potential for
    the number of refined values on a single engine. This is beneficial, since it uses the potential of
    the previous step as an initial guess of the next, making the iteration process converging much faster."""
    delta_V_G_range = refine_gridpoint(V_G, V_G_range, n_refine)
    init_guess_saving = []
    tf_pot_sol = []
    #times = []
    for i, V_G in enumerate(delta_V_G_range):
        if i == 0:
            #start_time = timeit.default_timer()
            init_guess_weight = 0.1*(V_G + 3*bound_conds.V_SC + 3*bound_conds.V_NL)
            init_guess = init_guess_weight*np.ones(discr_par.num_nodes)
            tf_pot_arr = tf_potential_anderson_iterative(V_G, init_guess, bound_conds, c, 
                                                         dim_par, discr_par, solver_kwargs)
            init_guess_saving.append(tf_pot_arr)
            #elapsed = timeit.default_timer() - start_time
            #times.append(elapsed)
        elif i > 0:
            #start_time = timeit.default_timer()
            init_guess = init_guess_saving[-1]
            tf_pot_arr = tf_potential_anderson_iterative(V_G, init_guess, bound_conds, c, 
                                                         dim_par, discr_par, solver_kwargs)
            init_guess_saving.append(tf_pot_arr)
            #elapsed = timeit.default_timer() - start_time
            #times.append(elapsed)
        tf_pot_sol.append(tf_pot_arr)
    return tf_pot_sol #, times
