"""
-----------------------------------------------------------------------
                         Hamiltonian Tools
-----------------------------------------------------------------------

(c) 2014 Anton Akhmerov (TU Delft).
See LICENSE.txt

This module allows to generate tight-binding onsite and hopping values
to be used in Kwant from a continuum Hamiltonian.
"""


import itertools
import re
from operator import mul
from functools import reduce
from collections import defaultdict
from math import factorial
import numpy as np
import sympy
import sympy.parsing
from sympy.printing.lambdarepr import LambdaPrinter
from sympy.utilities.lambdify import lambdastr
from sympy.physics.paulialgebra import Pauli, evaluate_pauli_product
from sympy.core.function import AppliedUndef
import tinyarray as ta


momenta = sympy.sympify('p_x, p_y, p_z')
p = sympy.Symbol('p')


class NumericPrinter(LambdaPrinter):
    def _print_ImaginaryUnit(self, expr):
        return "1.j"


# A simple container that may be used to store Hamiltonian parameters.
class SimpleNamespace(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


# Pauli matrices and Kronecker product.
kron = lambda *x: ta.array(reduce(np.kron, x, 1))
pauli = {'0': ta.array([[1, 0], [0, 1]], complex),
         'x': ta.array([[0, 1], [1, 0]], complex),
         'y': ta.array([[0, -1j], [1j, 0]], complex),
         'z': ta.array([[1, 0], [0, -1]], complex)}


_matrix_cache = {}


def make_tb_coefs():
    cache = {}

    def tb_coefs_1d(n):
        """Return tight-binding coefficients of p**n."""
        if n in cache:
            return cache[n]

        d = 1 + (n+1) // 2
        a = (1j * np.arange(1-d, d))**np.arange(2*d - 1).reshape(-1, 1)
        a = sympy.Matrix(a)
        a = a.inv().T
        result = list(factorial(n) * (a.row(n)))

        cache[n] = result
        return result
    return tb_coefs_1d

tb_coefs_1d = make_tb_coefs()
del make_tb_coefs


def tb_coefs(p_powers):
    """Calculate tight-binding coefficients of a monomial of momenta.

    Parameters:
    -----------
    p_powers : list of integers
       List with powers of momenta in each direction, e.g. `p_x**2 * p_y` would
       correspond to `(2, 1)`.

    Returns:
    --------
    coefficients : iterator
        An iterator of (vector, coefficient) tuples, where vector is an integer
        tuple corresponding to the hopping, and coefficient is a sympy
        constant expression (complex rational).
    """
    dir_coefs = [tb_coefs_1d(n) for n in p_powers]
    deltas = [list(range(-len(i)//2 + 1, len(i)//2 + 1)) for i in dir_coefs]
    positive = lambda i: i >= (sympy.Integer(0),) * len(i)
    hopping_vecs = (i for i in itertools.product(*deltas))
    hopping_coefs = (i for i in itertools.product(*dir_coefs))
    prod = lambda x, y=1: reduce(mul, x, y)
    result = ((i, prod(j)) for i, j in
              zip(hopping_vecs, hopping_coefs) if positive(i))
    return ((i, j) for i, j in result if j != 0)


def parse_ham(ham, pauli_names=()):
    """Translate a string defining the Hamiltonian into a sympy expression.

    This function is necessary only to take care of expressions that contain
    Pauli matrices that do not commute.

    Parameters:
    -----------
    ham : string
        A string containing an analytic expression.
    pauli_names : list of strings
        List of the names used to denote Pauli matrices.

    Returns:
    --------
    symbolic_hamiltonian : sympy.Expr
        The resulting Hamiltonian as a symbolic expression.

    Notes:
    ------
    The notation used here is that Pauli matrices in `ham` are specified as
    '<name>_<index>', where name is the name of the Pauli matrix (e.g. `sigma`,
    or `tau`), and index is one of 0, x, y, or z. All the products are
    interpreted as dot products, Products of Pauli matrices acting in different
    spaces are interpreted as direct products, and 1 is synonymous to identity
    matrix.
    """
    if pauli_names:
        pauli_match = '((?:' + "|".join(pauli_names) + ')_[xyz])'
        replacement = 'Symbol("\g<1>", commutative=False)'
        expr = sympy.sympify(re.sub(pauli_match, replacement, ham))
        idx = dict(list(zip('xyz', list(range(1, 4)))))
        terms = sympy.expand_mul(expr).as_ordered_terms()
        new_terms = []
        for term in terms:
            factors = term.as_ordered_factors()
            for name in pauli_names:
                to_contract = [i.name[-1] for i in factors if
                               hasattr(i, 'name') and i.name.startswith(name)]
                rest = [i for i in factors if not(hasattr(i, 'name')
                                                  and i.name.startswith(name))]
                new_factor = evaluate_pauli_product(sympy.Mul(*[Pauli(idx[i])
                                                    for i in to_contract]))
                reverse_idx = dict(list(zip(
                                   (Pauli(i) for i in range(1, 4)),
                                   (sympy.Symbol(name + '_' + i)
                                       for i in 'xyz'))))
                rest.append(new_factor.subs(reverse_idx))
                factors = rest
            new_terms.append(sympy.Mul(*factors))
        return sympy.Add(*new_terms)
    else:
        return sympy.sympify(ham)


def tb_ham(expr, dim=None, a=1.):
    """Discretize momentum operators in a continuum Hamiltonian.

    Parameters:
    -----------
    expr : sympy.Expr instance
        The continuum Hamiltonian containing momentum operators.
    dim : int or None
        Dimension of the Hamiltonian. If None, the dimensionality is inferred
        based on how many symbols out of (p_x, p_y, p_z) appear in the
        expression.
    a : float or sympy.Symbol
        Lattice constant. If it has a symbolic value, the lattice constant may
        be modified in the resulting Hamiltonian, and the namespace passed to
        the system as `args` should contain an attribute with the same name.

    Returns:
    --------
    hoppings : dict
        A dictionary where keys are integer tuples corresponding to lattice
        vectors, and values are the sympy expressions for the corresponding
        hoppings. Out of two hoppings in the opposite directions only the
        lexicographically larger one is present. The hopping consisting out of
        only zeros is the onsite Hamiltonian.
    dim : int
        The resulting system dimensionality, useful when dimension is
        inferred.
    """
    available_momenta = [i for i in momenta if expr.has(i)]

    # Determine the dimensionality.
    if dim is None:
        if not available_momenta:
            raise RuntimeError("Failed to defermine dimensionality,"
                               "specify the value of dim.")
        dim = len(available_momenta)
    elif dim == int(dim) and 1 <= dim <= 3:
        extra_momenta = [i for i in available_momenta
                         if i not in momenta[:dim]]
        if extra_momenta:
            singular = 's' * (len(extra_momenta) <= 1)
            msg = "dim is set to {0}, but {1} appear {2} in the Hamiltonian."
            momenta_names = ', '.join([i.name for i in extra_momenta])
            raise RuntimeError(msg.format(dim, momenta_names, singular))
        available_momenta = momenta[:dim]
    else:
        raise ValueError("dim should be an integer between 1 and 3.")

    expr = expr.replace(p, sympy.sqrt(sum(p**2 for p in available_momenta)))
    expr = sympy.polys.Poly(expr, gens=available_momenta)
    monomials = ((power, coef)
            for power, coef in zip(expr.monoms(), expr.coeffs()))
    hoppings = defaultdict(lambda: 0)
    for p_powers, value in monomials:
        for vec, coef in tb_coefs(p_powers):
            hoppings[vec] += (value * coef / a**sum(p_powers)).n()
    return dict(hoppings), dim


def make_return_string(expr, pauli_names):
    """Process a sympy expression into an evaluatable Python return statement.

    Parameters:
    -----------
    expr : sympy.Expr instance
        The expression for the Hamiltonian. It shouldn't contain any
        mathematical functions that are not in the standard library.
    pauli_names : list of strings
        List of the names used to denote Pauli matrices. The order in which the
        names appear in this list is the same as the one to be used for the
        Kronecker product.

    Returns:
    --------
    output : string
        A return string that can be used to assemble a Kwant value function.
    free_symbols : list of sympy.Symbol instances
        All the undefined variables that appear in the expression.
    """
    full_pauli_names = ['_'.join(ij)
                        for ij in itertools.product(pauli_names, 'xyz')]
    pauli_symbols = [sympy.Symbol(i) for i in full_pauli_names]

    def monomial_to_matrix(coefs):
        coefs = [iter(coefs)] * 3
        letters = []
        for i in zip(*coefs):
            if any(i):
                letters.append('xyz'[i.index(1)])
            else:
                letters.append('0')
        return ''.join(letters), kron(*[pauli[letter] for letter in letters])

    global _matrix_cache
    pauli_polynomial = sympy.polys.Poly(expr, gens=pauli_symbols)
    terms = []

    for monom, coef in zip(pauli_polynomial.monoms(),
                           pauli_polynomial.coeffs()):
        name, matrix = monomial_to_matrix(monom)
        _matrix_cache[name] = matrix
        term = ('_matrix_cache["{0}"] * '.format(name) +
                lambdastr((), coef.simplify(),
                printer=NumericPrinter)[len('lambda : '):])
        terms.append(term)
    free_symbols = {i for i in expr.free_symbols
                    if not i.name in full_pauli_names}

    return 'return ' + ' + '.join(terms), free_symbols


def assign_symbols(symbol_assignments, free_symbols, dim):
    """Generate a series of assignments defining a set symbols.

    Parameters:
    -----------
    symbol_assignments : list of strings
        A list of strings that specify the values of parameters.
    free_symbols : list of sympy.Symbol instances
        The initial set of free variables (see notes).
    dim : int
        Dimension of the system.

    Returns:
    --------
    assignments : list of strings
        List of lines used for including in a function.

    Notes:
    ------
    The resulting lines begin with a line of form
    `A, B, C = p.A, p.B, p.C`
    where A, B, C are all the free symbols plus the symbols that appear on the
    right hand side of assignments, and do not appear on the left hand
    side. Then, if coordinates are used anywhere in the replacements a line of
    the form `x, y, z = site.pos` follows. Finally, all input lines follow. The
    main use of this function is to allow to specify the functional form of the
    Hamiltonian parameters through assignments like:
    `mu = mu_averae + barrier(x, y) + disorder(site, salt)`
    Then the namespace specifying the Hamiltonian parameters should contain a
    function of two arguments `barrier` and another function `disorder` as its
    attributes.
    """
    symbols = set(free_symbols)
    coord_names = 'xyz'[:dim]
    lines = list(symbol_assignments)
    for assignment in reversed(symbol_assignments):
        lhs, rhs = assignment.split('=')
        try:
            symbols.remove(sympy.sympify(lhs))
        except KeyError:  # Unused parameter.
            lines.remove(assignment)
            continue
        rhs = sympy.sympify(rhs)
        func_symbols = {sympy.Symbol(i.func.__name__) for i in
                        rhs.atoms(AppliedUndef)}
        symbols = symbols | rhs.free_symbols | func_symbols
    if any(i.name in coord_names for i in symbols):
        lines.insert(0, ', '.join(coord_names) + ' = site.pos')
    symbols -= {sympy.Symbol(i) for i in coord_names}
    symbols -= {sympy.Symbol(name) for name in ('site', 'site1', 'site2')}
    symb_names = [i.name for i in symbols]
    if symb_names:
        lines.insert(0, ', '.join(symb_names) + ' = p.' +
                     ', p.'.join(symb_names))
    return lines


def value_function(content, name='_anonymous_func',
                   onsite=True, verbose=False):
    """Generate a Kwant value function from a list of lines containing its body.

    Parameters:
    -----------
    content : list of lines
        Lines forming the body of the function.
    name : string
        Function name (not important).
    onsite : bool
        If True, the function call signature will be `f(site, p)`, otherwise
        `f(site1, site2, p)`.
    verbose : bool
        Whether the function bodies should be printed.

    Returns:
    --------
    f : function
        The function defined in a namespace containing only the cached values
        of Pauli matrices.
    """
    if not content[-1].startswith('return'):
        raise ValueError('The function does not end with a return statement')
    separator = '\n' + 4 * ' '
    site_string = 'site' if onsite else 'site1, site2'
    header = 'def {0}({1}, p):'.format(name, site_string)
    func_code = separator.join([header] + list(content))
    namespace = {'_matrix_cache': _matrix_cache}
    if verbose:
        print(func_code)
    exec(func_code, namespace)
    return namespace[name]


def make_tb_system(model, symbol_assignments, pauli_names=(), dim=None, a=1,
                   verbose=False):
    """A convenience function that generates Kwant system values from a model.

    This function can be replaced by subsequent calls to other functions
    defined in this module.

    Parameters:
    -----------
    model : string
        Symbolic definition of the Hamiltonian.
    symbol_assignments : list of strings
        A list of strings that specify the values of parameters.
    pauli_names : list of strings
        List of the names used to denote Pauli matrices. The order in which the
        names appear in this list is the same as the one to be used for the
        Kronecker product.
    dim : int or None
        Dimension of the Hamiltonian. If None, the dimensionality is inferred
        based on how many symbols out of (p_x, p_y, p_z) appear in the
        expression.
    a : float or sympy.Symbol
        Lattice constant. If it has a symbolic value, the lattice constant may
        be modified in the resulting Hamiltonian, and the namespace passed to
        the system as `args` should contain an attribute with the same name.
    verbose : bool
        Whether the bodies of the resulting function should be printed.

    Returns:
    onsite : function
        The value of the onsite Hamiltonian.
    hoppings : dict
        A dictionary with keys being tuples of the lattice hopping, and values
        the corresponding value functions.
    """
    def make_func(ham, onsite):
        return_string, symbols = make_return_string(ham, pauli_names)
        lines = assign_symbols(symbol_assignments, symbols, dim)
        lines.append(return_string)
        return value_function(lines, verbose=verbose, onsite=onsite)

    expr = parse_ham(model, pauli_names)
    hop_dict, dim = tb_ham(expr, dim, a)
    return make_func(hop_dict.pop((0,) * dim), True), {k: make_func(v, False)
                                                       for k, v in
                                                       list(hop_dict.items())}
